// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::mem;

use super::{World, EffectController, super::SpecialArea, super::powerup::{Effect, active_effects::EffectInfo}};

use crate::utils::maths::{RectCorners, CollidingRect};
use crate::net::server::config::{EventConfig, EffectZone as ConfigEffectZone, EffectZoneDefaultColourConfig};

pub struct EffectZones(Box<[EffectZone]>);

struct EffectZone {
	rect: RectCorners,
	effect: Effect,
	info: EffectInfo,
}

impl EffectZones {
	pub fn new(zones: Box<[ConfigEffectZone]>, event_config: &EventConfig, default_colours: EffectZoneDefaultColourConfig, special_areas: &mut Vec<SpecialArea>) -> EffectZones {
		EffectZones(zones.into_vec().into_iter().map(|mut zone| {
			for i in 0..2 {
				if zone.p0[i] > zone.p1[i] {
					mem::swap(&mut zone.p0[i], &mut zone.p1[i]);
				}
			}

			let colour = zone.colour.unwrap_or(match zone.effect {
				Effect::Speed => default_colours.speed,
				Effect::Reload => default_colours.reload,
				Effect::Regeneration => default_colours.regen,
				Effect::Damage => default_colours.damage,
				Effect::Forcefield => default_colours.forcefield,
			});

			special_areas.push(SpecialArea { pos: zone.p0, size: zone.p1 - zone.p0, colour });

			EffectZone {
				rect: RectCorners(zone.p0, zone.p1),
				effect: zone.effect,
				info: EffectInfo { time: zone.time, power: zone.power.unwrap_or_else(|| event_config.get_effect_power(zone.effect)) },
			}
		}).collect::<Vec<_>>().into_boxed_slice())
	}

	pub fn update(&self, world: &World, effect_controller: &mut EffectController) {
		for (&id, player) in &world.players {
			for zone in &self.0 {
				if player.colliding_rect(&zone.rect) {
					effect_controller.in_zone(id, zone.effect, zone.info);
				}
			}
		}
	}
}
