// Copyright Marcus Del Favero 2025
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::{rc::Rc, cell::RefCell};

use glium::{Display, Frame, DrawParameters, StencilTest, StencilOperation};
use glutin::surface::WindowSurface;
use glam::Mat4;

use super::{
	World, SmoothWorld, border_renderer::BorderRenderer, player::{PlayerRenderer,
	forcefield::{ForcefieldRenderer, RenderedForcefield}},
	bullet::BulletRenderer, ammo_crate::AmmoCrateRenderer,
	powerup::PowerupRenderer, special_area::SpecialAreaRenderer,
};

use crate::game::{config::Config, filesystem::Filesystem, player_gfx::PlayerGfx};
use crate::playing::{model::Model, view::Camera};
use crate::blocks::{Blocks, BlocksRenderer};
use crate::utils::blending::{ALPHA_BLENDING_FUNC, ADDITIVE_BLENDING_FUNC};

const MIN_TINT: f32 = 0.4375; // Brightness of the initial tint
const TINT_TIME: f32 = 0.15625;

pub struct Renderer {
	border: BorderRenderer,
	players: PlayerRenderer,
	bullets: BulletRenderer,
	ammo_crates: AmmoCrateRenderer,
	powerups: PowerupRenderer,
	forcefields: ForcefieldRenderer,
	blocks: BlocksRenderer,
	special_areas: SpecialAreaRenderer,
}

impl Renderer {
	pub fn new(display: &Display<WindowSurface>, fs: &Filesystem, config: Rc<RefCell<Config>>, player_gfx: Rc<PlayerGfx>) -> Renderer {
		Renderer {
			border: BorderRenderer::new(display, fs),
			players: PlayerRenderer::new(display, player_gfx),
			bullets: BulletRenderer::new(display, fs),
			ammo_crates: AmmoCrateRenderer::new(display, fs),
			powerups: PowerupRenderer::new(display, fs),
			forcefields: ForcefieldRenderer::new(config, display, fs),
			blocks: BlocksRenderer::new(display, fs),
			special_areas: SpecialAreaRenderer::new(display, fs),
		}
	}

	pub fn reset(&mut self, display: &Display<WindowSurface>, model: &Model) {
		self.border.reset(model.get_world());
		self.blocks.reset(model.get_blocks());
		self.special_areas.update(&model.get_world().special_areas, display);
	}

	pub fn render_border(&self, frame: &mut Frame, params: &DrawParameters, matrix: &Mat4) {
		self.border.render(frame, params, matrix);
	}

	pub fn render_special_areas(&mut self, display: &Display<WindowSurface>, frame: &mut Frame, params: &mut DrawParameters, blocks: &Blocks, camera: &Camera, matrix: &Mat4) {
		/*
		 * Uses the stencil test to avoid special areas from being visible under
		 * blocks. This looks pretty ugly as blocks are now partially transparent.
		 */

		// Disables rendering blocks
		params.color_mask = [false; 4].into();

		// Sets the stencil buffer value to 1
		params.stencil.reference_value_counter_clockwise = 1;
		params.stencil.reference_value_clockwise = 1;
		params.stencil.depth_pass_operation_counter_clockwise = StencilOperation::Replace;
		params.stencil.depth_pass_operation_clockwise = StencilOperation::Replace;

		self.blocks.render(display, frame, params, blocks, camera, false);

		// Reverts all settings
		params.color_mask = [true; 4].into();
		params.stencil.reference_value_counter_clockwise = 0;
		params.stencil.reference_value_clockwise = 0;
		params.stencil.depth_pass_operation_counter_clockwise = StencilOperation::Keep;
		params.stencil.depth_pass_operation_clockwise = StencilOperation::Keep;

		// Keeps fragments only if they are equal to zero (the reference value), which means not part of the blocks (previously set to 1)
		params.stencil.test_counter_clockwise = StencilTest::IfEqual { mask: u32::MAX };
		params.stencil.test_clockwise = StencilTest::IfEqual { mask: u32::MAX };

		self.special_areas.render(frame, params, matrix);

		// Reverts the settings
		params.stencil.test_counter_clockwise = StencilTest::AlwaysPass;
		params.stencil.test_clockwise = StencilTest::AlwaysPass;
	}

	#[allow(clippy::too_many_arguments)]
	pub fn render(&mut self, dt: f32,
		display: &Display<WindowSurface>, frame: &mut Frame, params: &mut DrawParameters,
		time_diff: f32, world: &World, smooth_world: &mut SmoothWorld, blocks: &Blocks, forcefields: &[RenderedForcefield], camera: &Camera,
	) {
		let matrix = camera.get_matrix();
		let time_shift = smooth_world.get_total_time_shift() + time_diff;

		self.ammo_crates.render(world.get_ammo_crates(), display, frame, params, matrix, time_shift);
		params.blend.color = ADDITIVE_BLENDING_FUNC; // The powerups are basically "lightning balls" so additive blending can be used (also makes texture generation easier)
		self.powerups.render(world.get_powerups(), display, frame, params, matrix, time_shift);
		params.blend.color = ALPHA_BLENDING_FUNC;

		self.bullets.render(smooth_world.get_bullets(), forcefields, display, frame, params, matrix, time_diff);

		let followed_id = smooth_world.followed_player_id();
		let mut followed_info = None;
		for (id, player) in smooth_world.player_iter_mut() {
			let tint = if player.last_hurt < TINT_TIME {
				(player.last_hurt / TINT_TIME) * (1.0 - MIN_TINT) + MIN_TINT
			} else {
				1.0
			};

			player.hbar.update(&player.player, dt);
			if Some(id) == followed_id {
				followed_info = Some((tint, time_diff));
			} else {
				self.players.push(player, tint, time_diff);
			}
			player.update_track(dt, time_diff);
			player.last_hurt += dt; // Want to update the tint after rendering so the darkest tint is rendered once
		}

		// Renders the followed player above all other players
		if let Some(((tint, time_diff), (_id, player))) = followed_info.zip(smooth_world.followed_player()) {
			self.players.push(player, tint, time_diff);
		}

		self.players.render(display, frame, params, matrix);

		self.blocks.render(display, frame, params, blocks, camera, true);

		if !forcefields.is_empty() {
			self.forcefields.update(dt);
			self.forcefields.render(forcefields, display, frame, params, matrix);
		}
	}
}
